@@ -5,11 +5,13 @@ from torch import nn, einsum
 from torchaudio.transforms import Spectrogram, TimeStretch, FrequencyMasking, TimeMasking
 
 from audiolm_pytorch import AudioLM
+from audiolm_pytorch.utils import AudioConditionerBase
 
 from x_clip.tokenizer import tokenizer
 from vector_quantize_pytorch import ResidualVQ
 
 from einops import rearrange, repeat, reduce, pack, unpack
+from einops.layers.torch import Rearrange
 
 from beartype.typing import List, Optional, Tuple
 from beartype import beartype
@@ -25,6 +27,9 @@ def default(val, d):
 def round_down_nearest_multiple(n, divisor):
     return n // divisor * divisor
 
+def Sequential(*modules):
+    return nn.Sequential(*filter(exists, modules))
+
 # tensor functions
 
 def log(t, eps = 1e-20):
@@ -188,6 +193,27 @@ class Transformer(nn.Module):
 
         return x
 
+# Patch Dropout - https://arxiv.org/abs/2208.07220
+
+class PatchDropout(nn.Module):
+    def __init__(self, prob):
+        super().__init__()
+        assert 0 <= prob < 1.
+        self.prob = prob
+
+    def forward(self, x, force_keep_all = False):
+        if not self.training or self.prob == 0. or force_keep_all:
+            return x
+
+        b, n, _, device = *x.shape, x.device
+
+        batch_indices = torch.arange(b, device = device)
+        batch_indices = rearrange(batch_indices, '... -> ... 1')
+        num_patches_keep = max(1, int(n * (1 - self.prob)))
+        patch_indices_keep = torch.randn(b, n, device = device).topk(num_patches_keep, dim = -1).indices
+
+        return x[batch_indices, patch_indices_keep]
+
 # Audio Spectrogram Transformer - https://arxiv.org/abs/2104.01778
 
 def pair(t):
@@ -213,14 +239,22 @@ class AudioSpectrogramTransformer(nn.Module):
         spec_pad_mode = 'reflect',
         spec_aug_stretch_factor = 0.8,
         spec_aug_freq_mask = 80,
-        spec_aug_time_mask = 80
-
+        spec_aug_time_mask = 80,
+        dual_patchnorm = True,
+        patch_dropout_prob = 0.5
     ):
         super().__init__()
         self.dim = dim
 
         self.patch_size = pair(patch_size)
-        self.to_patch_tokens = nn.Conv2d(self.patch_size[0] * self.patch_size[1], dim, 1)
+        patch_input_dim = self.patch_size[0] * self.patch_size[1]
+
+        self.to_patch_tokens = Sequential(
+            Rearrange('b (h p1) (w p2) -> b h w (p1 p2)', p1 = self.patch_size[0], p2 = self.patch_size[1]),
+            nn.LayerNorm(patch_input_dim) if dual_patchnorm else None,
+            nn.Linear(patch_input_dim, dim),
+            nn.LayerNorm(dim) if dual_patchnorm else None
+        )
 
         self.spec = Spectrogram(
             n_fft = spec_n_fft,
@@ -252,6 +286,8 @@ class AudioSpectrogramTransformer(nn.Module):
 
         self.norm = LayerNorm(dim)
 
+        self.patch_dropout = PatchDropout(patch_dropout_prob)
+
     def forward(self, x):
         x = self.spec(x)
 
@@ -272,18 +308,18 @@ class AudioSpectrogramTransformer(nn.Module):
 
         # to patches
 
-        x = rearrange(x, 'b (h p1) (w p2) -> b (p1 p2) h w', p1 = patch_height, p2 = patch_width)
         x = self.to_patch_tokens(x)
 
         # 2d sinusoidal positional embedding
 
-        x = rearrange(x, 'b c h w -> b h w c')
         x = x + posemb_sincos_2d(x)
 
         # attention, what else
 
         x = rearrange(x, 'b ... c -> b (...) c')
 
+        x = self.patch_dropout(x)
+
         x = self.transformer(x)
 
         # final global average and norm (most recent papers show this is superior to CLS token)
@@ -448,7 +484,7 @@ class MuLaN(nn.Module):
 # music lm
 
 @beartype
-class MuLaNEmbedQuantizer(nn.Module):
+class MuLaNEmbedQuantizer(AudioConditionerBase):
     def __init__(
         self,
         mulan: MuLaN,
@@ -494,6 +530,9 @@ class MuLaNEmbedQuantizer(nn.Module):
 
         self.set_default_namespace(namespaces[0])
 
+    def parameters(self):
+        return self.cond_embeddings.parameters()
+
     def set_default_namespace(self, namespace):
         self._default_namespace = namespace
 
@@ -537,6 +576,8 @@ class MusicLM(nn.Module):
         mulan_embed_quantizer: MuLaNEmbedQuantizer
     ):
         super().__init__()
+        assert not exists(audio_lm.audio_conditioner), 'mulan must not have been passed into AudioLM. it will be managed externally now, embedding the text into the joint embedding space for text-to-audio synthesis'
+
         self.mulan_embed_quantizer = mulan_embed_quantizer
         self.audio_lm = audio_lm
 
